//
//  ScoreView.swift
//  TERANE Microtonal Score Editor
//
//  Created by Ozan Yarman on 20.10.2024.
//  Copyright © 2024 Ozan Yarman. All rights reserved.
//

import Cocoa
import Foundation
import CoreGraphics  // For CGPoint

extension CGPoint {
    /// Calculate the Euclidean distance between two points.
    func distance(to point: CGPoint) -> CGFloat {
        let dx = self.x - point.x
        let dy = self.y - point.y
        return sqrt(dx * dx + dy * dy)
    }
}

protocol ScoreViewDelegate: AnyObject {
    func playNoteFromScoreView(_ note: Note)
}

class ScoreView: NSView {
    weak var delegate: ScoreViewDelegate?
    
    // Array to store the Note objects, not just points
    var notes: [Note] = []
    
    // This will be set by the ViewController to hold the selected note and its offset
    var selectedNote: Note?
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)
        
        // Add a background color
        NSColor.white.setFill()
        NSBezierPath(rect: dirtyRect).fill()
        
        // Set the color for the staff lines
        NSColor.black.set()
        
        // Draw the staff lines
        let staffLineSpacing: CGFloat = 10.0
        let staffYStart: CGFloat = bounds.height / 2 - 2 * staffLineSpacing
        
        let path = NSBezierPath()
        for i in 0..<5 {
            let yPosition = staffYStart + CGFloat(i) * staffLineSpacing
            path.move(to: CGPoint(x: 50, y: yPosition)) // Start at x: 50 to leave space for the clef
            path.line(to: CGPoint(x: bounds.width - 10, y: yPosition))
        }
        path.lineWidth = 2.0
        path.stroke()
        
        // Draw the G clef using the Bravura font
        drawGClef(at: CGPoint(x: 20, y: staffYStart - staffLineSpacing))
        
        // Draw the notes
        for note in notes {
            drawNoteAt(position: note.position)
        }
    }
    
    // Function to draw the G clef
    func drawGClef(at position: CGPoint) {
        let clefFont = NSFont(name: "Bravura", size: 40) // Use the Bravura font
        let clefAttributes: [NSAttributedString.Key: Any] = [
            .font: clefFont!,
            .foregroundColor: NSColor.black
        ]
        
        let gClef = "\u{E050}" // G clef symbol from Bravura font
        gClef.draw(at: position, withAttributes: clefAttributes)
    }
    
    // Function to draw a note at a specific position
    func drawNoteAt(position: CGPoint) {
        let notePath = NSBezierPath(ovalIn: CGRect(x: position.x - 5, y: position.y - 5, width: 10, height: 10))
        notePath.fill()
    }
    
    // This function will be called when the user clicks the mouse
    override func mouseDown(with event: NSEvent) {
        let location = convert(event.locationInWindow, from: nil)
        
        // Check if the user clicked on an existing note
        if let selectedNoteIndex = notes.firstIndex(where: { $0.position.distance(to: location) < 10 }) {
            let existingNote = notes[selectedNoteIndex]
            print("Selected existing note: \(existingNote.pitch)\(existingNote.octave), Frequency: \(existingNote.frequency) Hz")
        } else {
            let (pitch, octave) = mapLocationToNotePitchAndOctave(location: location)
            let baseFrequency = baseFrequencies[pitch] ?? 261.63
            let centOffset = selectedNote?.centOffset ?? 0.0
            
            let newNote = Note(pitch: pitch, octave: octave, baseFrequency: baseFrequency, centOffset: centOffset, position: location)
            notes.append(newNote)
            setNeedsDisplay(bounds)
            
            print("Added new note: \(newNote.pitch)\(newNote.octave), Frequency: \(newNote.frequency) Hz")
            
            // Use the delegate to trigger sound in ViewController
            delegate?.playNoteFromScoreView(newNote)
        }
    }

    // Method to map y-coordinate to note pitch and octave
    func mapLocationToNotePitchAndOctave(location: CGPoint) -> (String, Int) {
        // (Same as previously explained)
        let staffSpacing: CGFloat = 20.0
        let y = location.y
        switch y {
        case 0..<staffSpacing: return ("C", 4)
        case staffSpacing..<(2 * staffSpacing): return ("D", 4)
        case (2 * staffSpacing)..<(3 * staffSpacing): return ("E", 4)
        case (3 * staffSpacing)..<(4 * staffSpacing): return ("F", 4)
        case (4 * staffSpacing)..<(5 * staffSpacing): return ("G", 4)
        case (5 * staffSpacing)..<(6 * staffSpacing): return ("A", 4)
        case (6 * staffSpacing)..<(7 * staffSpacing): return ("B", 4)
        case (7 * staffSpacing)..<(8 * staffSpacing): return ("C", 5)
        default: return ("C", 4)
        }
    }
    
    // This function will be called by the ViewController when the user selects a note
    func updateSelectedNote(pitch: String, octave: Int, centOffset: Float) {
        // Get the base frequency for the selected pitch
        guard let baseFrequency = baseFrequencies[pitch] else {
            print("Invalid pitch selected")
            return
        }
        
        // Update the selected note
        selectedNote = Note(pitch: pitch,
                            octave: octave,
                            baseFrequency: baseFrequency,
                            centOffset: centOffset,
                            position: CGPoint.zero) // Position will be defined on mouse click
    }
        
}
