//
//  ViewController.swift
//  TERANE Microtonal Score Editor
//
//  Created by Ozan Yarman on 19.10.2024.
//  Copyright © 2024 Ozan Yarman. All rights reserved.
//

import Cocoa
import Foundation
import AVFoundation
import AudioToolbox

// MARK: - DynamicLevel Enum

enum DynamicLevel: String, CaseIterable {
    case pppp = "pppp"
    case ppp = "ppp"
    case pp = "pp"
    case p = "p"
    case mp = "mp"
    case mf = "mf"
    case f = "f"
    case ff = "ff"
    case fff = "fff"
    case ffff = "ffff"
    case fp = "fp"
    case ffp = "ffp"
    case ffpp = "ffpp"
    case fffppp = "fffppp"
    case ffffpppp = "ffffpppp"
    case sp = "sp"
    case spp = "spp"
    case sppp = "sppp"
    case spppp = "spppp"
    case sf = "sf"
    case sff = "sff"
    case sfff = "sfff"
    case sffff = "sffff"
    case sfz = "sfz"
    case sffz = "sffz"
    case sfffz = "sfffz"
    case sffffz = "sffffz"
    case rfz = "rfz"
    case rffz = "rffz"
    case rfffz = "rfffz"
    case rffffz = "rffffz"
    case sfp = "sfp"
    case sfpp = "sfpp"
    case sfppp = "sfppp"
    case sfpppp = "sfpppp"
    case cresc = "cresc."
    case crescendo = "crescendo"
    case crescPocoPoco = "cresc. poco a poco"
    case crescendoPocoPoco = "crescendo poco a poco"
    case decresc = "decresc."
    case decrescendo = "decrescendo"
    case decrescPocoPoco = "decresc. poco a poco"
    case decrescendoPocoPoco = "decrescendo poco a poco"
    case diminuendofullabbrev = "dim."
    case diminuendohalfabbrev = "dimin."
    case diminuendo = "diminuendo"
    case diminuendoPocoPoco = "diminuendo poco a poco"
    case smz = "smz."
    case smorzando = "smorzando"
    case morendo = "morendo"
    case perdendosi = "perdendosi"
}

// Public Struct to represent a note
public struct Note {
    var pitch: String  // The letter name of the note (C, D, E, etc.)
    var octave: Int    // The octave number (e.g., 4 for C4)
    var baseFrequency: Float  // The base frequency of the note in Hz
    var centOffset: Float     // The microtonal adjustment in cents
    
    // Computed property to calculate the final frequency with cent adjustment
    var frequency: Float {
        return baseFrequency * pow(2.0, centOffset / 1200.0)
    }
    
    var position: CGPoint  // The position of the note on the staff
}

// Dictionary of base frequencies for natural notes (4th octave)
let baseFrequencies: [String: Float] = [
    "C": 261.63,   // C4
    "D": 293.66,   // D4
    "E": 329.63,   // E4
    "F": 349.23,   // F4
    "G": 392.00,   // G4
    "A": 440.00,   // A4
    "B": 493.88    // B4
]

// Define an error type for invalid pitches
enum NoteError: Error {
    case invalidPitch
}

class ViewController: NSViewController, ScoreViewDelegate {

    // Outlet for ScoreView (the custom view)
    @IBOutlet weak var scoreView: ScoreView!
    
    // Outlet for dynamic picker (drop-down list)
    @IBOutlet weak var dynamicPicker: NSPopUpButton!
    @IBOutlet weak var selectedDynamicLabel: NSTextField!

    // Action triggered when the popup button selection changes
    @IBAction func dynamicPickerChanged(_ sender: NSPopUpButton) {
        updateSelectedDynamicLabel()
    }
    
    @IBOutlet weak var C_Slider: NSSlider!
    @IBOutlet weak var D_Slider: NSSlider!
    @IBOutlet weak var E_Slider: NSSlider!
    @IBOutlet weak var F_Slider: NSSlider!
    @IBOutlet weak var G_Slider: NSSlider!
    @IBOutlet weak var A_Slider: NSSlider!
    @IBOutlet weak var B_Slider: NSSlider!
    
    @IBAction func C_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "C", centOffset: sender.floatValue)
    }

    @IBAction func D_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "D", centOffset: sender.floatValue)
    }

    @IBAction func E_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "E", centOffset: sender.floatValue)
    }

    @IBAction func F_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "F", centOffset: sender.floatValue)
    }

    @IBAction func G_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "G", centOffset: sender.floatValue)
    }

    @IBAction func A_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "A", centOffset: sender.floatValue)
    }

    @IBAction func B_SliderChanged(_ sender: NSSlider) {
        updateNoteCalibration(pitch: "B", centOffset: sender.floatValue)
    }
    
    var noteCalibrations: [String: Float] = [
        "C": 0.0,
        "D": 0.0,
        "E": 0.0,
        "F": 0.0,
        "G": 0.0,
        "A": 0.0,
        "B": 0.0
    ]
    
    func updateNoteCalibration(pitch: String, centOffset: Float) {
        noteCalibrations[pitch] = centOffset
        print("Updated \(pitch) calibration to \(centOffset) cents")
    }

    // Array of dynamic levels
    let dynamicLevels = DynamicLevel.allCases
    
    var audioEngine: AVAudioEngine!
    var sampler: AVAudioUnitSampler!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        setupAudioEngine()
        
        // Populate the picker with the dynamic levels
        dynamicPicker.removeAllItems()
        for dynamic in dynamicLevels {
            dynamicPicker.addItem(withTitle: dynamic.rawValue)
        }
        
        // Set default dynamic
        dynamicPicker.selectItem(at: 0)
        updateSelectedDynamicLabel()
    }
    
    func setupAudioEngine() {
        audioEngine = AVAudioEngine()
        sampler = AVAudioUnitSampler()
        
        audioEngine.attach(sampler)
        audioEngine.connect(sampler, to: audioEngine.mainMixerNode, format: nil)
        
        do {
            try audioEngine.start()
            print("Audio engine started successfully.")
        } catch {
            print("Failed to start audio engine: \(error.localizedDescription)")
        }
    }
    
    func loadSoundFont() {
        let soundFontPath = "/Library/Audio/Sounds/Banks/Final.sf2"
        let soundFontURL = URL(fileURLWithPath: soundFontPath)

        do {
            let bank: UInt8 = 0x79
            let patch: UInt8 = 0
            try sampler.loadSoundBankInstrument(at: soundFontURL, program: patch, bankMSB: bank, bankLSB: 0)
        } catch {
            print("Failed to load SoundFont: \(error.localizedDescription)")
        }
    }
    
    func playNote(note: Note) {
        do {
            let midiNote = try pitchToMIDINoteNumber(pitch: note.pitch, octave: note.octave)
            sampler.startNote(UInt8(midiNote), withVelocity: 127, onChannel: 0)

            DispatchQueue.main.asyncAfter(deadline: .now() + 1.0) {
                self.sampler.stopNote(UInt8(midiNote), onChannel: 0)
            }
        } catch {
            print("Error playing note: \(error.localizedDescription)")
        }
    }

    // Conform to the ScoreViewDelegate protocol
    func playNoteFromScoreView(_ note: Note) {
        playNote(note: note)
    }
    
    func pitchToMIDINoteNumber(pitch: String, octave: Int) throws -> Int {
        let noteToMIDIMapping: [String: Int] = [
            "C": 0, "C#": 1, "D": 2, "D#": 3, "E": 4, "F": 5, "F#": 6,
            "G": 7, "G#": 8, "A": 9, "A#": 10, "B": 11
        ]
        
        guard let baseMIDINote = noteToMIDIMapping[pitch] else {
            throw NoteError.invalidPitch
        }
        
        let midiNoteNumber = baseMIDINote + (octave + 1) * 12
        return midiNoteNumber
    }

    // Update the dynamic label based on the current selection
    func updateSelectedDynamicLabel() {
        if let selectedTitle = dynamicPicker.titleOfSelectedItem {
            selectedDynamicLabel.stringValue = "Selected Dynamic: \(selectedTitle)"
        }
    }
    
    @IBAction func noteSelectionChanged(_ sender: NSPopUpButton) {
        let selectedTitle = sender.titleOfSelectedItem ?? "C"
        
        let pitch: String
        let centOffset: Float
        
        switch selectedTitle {
        case "C#":
            pitch = "C"
            centOffset = 100.0
        case "D#":
            pitch = "D"
            centOffset = 100.0
        case "F#":
            pitch = "F"
            centOffset = 100.0
        case "G#":
            pitch = "G"
            centOffset = 100.0
        case "A#":
            pitch = "A"
            centOffset = 100.0
        default:
            pitch = selectedTitle
            centOffset = noteCalibrations[pitch] ?? 0.0
        }
        
        scoreView.updateSelectedNote(pitch: pitch, octave: 4, centOffset: centOffset)
        
        // Set the delegate for the ScoreView
        scoreView.delegate = self
    }
}
